@@ -1,24 +1,12 @@
 '''
 ext_Experiment is the container for whole project data.
+sample_registry is the dict to track sample information and status,
+outside class to facilitate multiprocessing.
 Heavy lifting is in constructors.CompositeMap, 
     which contains MassGrid for correspondence, and FeatureList from feature/peak detection.
 Annotation is facilitated by jms-metabolite-services, mass2chem. 
-
-
-
-        if sample_N <= 10:
-            Start by matching anchor pairs, then work thru remaining traces.
-            1. create a reference based on anchor pairs
-            2. align each sample to the reference_anchor_pairs
-        elif sample_N <= 1000:     # more samples use a different method, as peak density will be apparent in more samples.
-            single batch build
-        else:                    # even more samples should be split to batches for performance reasons
-            multiple batch build
-
-
 '''
 import time
-import random
 import multiprocessing as mp
 
 import pymzml
@@ -30,11 +18,46 @@ from .mass_functions import *
 # -----------------------------------------------------------------------------
 # main workflow for `process`
 # -----------------------------------------------------------------------------
+def process_project(list_input_files, parameters):
+    '''
+    list_input_files: mzML files.
+    parameters: dictionary of parameters.
+
+    Will adjust disk writing according to database_mode
+    '''
+    sample_registry = register_samples(list_input_files)
+    if parameters['database_mode'] == 'auto':
+        if len(list_input_files) <= parameters['project_sample_number_small']:
+            parameters['database_mode'] = 'memory'
+        else:
+            parameters['database_mode'] = 'ondisk'        # yet to implement mongo etc
+
+    # time_stamp is `month daay hour minute second``
+    time_stamp = [str(x) for x in time.localtime()[1:6]]
+    parameters['time_stamp'] = ':'.join(time_stamp)
+    time_stamp = ''.join(time_stamp)
+    # if parameters['database_mode'] == 'ondisk':
+    create_export_folders(parameters, time_stamp)
+        
+    # samples are processed to mass tracks (EICs) here
+    shared_dict = batch_EIC_from_samples_(sample_registry, parameters)
+    for sid, sam in sample_registry.items():
+        sam['status:mzml_parsing'], sam['status:eic'], sam['data_location'
+            ], sam['max_scan_number'], sam['list_scan_numbers'], sam['list_retention_time'
+            ], sam['track_mzs'
+            ], sam['number_anchor_mz_pairs'], sam['anchor_mz_pairs'
+            ], sam['sample_data'] = shared_dict[sid]
+
+        sam['name'] = os.path.basename(sam['input_file']).replace('.mzML', '')
+    
+    EE = ext_Experiment(sample_registry, parameters)
+    EE.process_all()
+    EE.export_all()
+
 
 def read_project_dir(directory, file_pattern='.mzML'):
     '''
-    This reads centroided LC-MS files.
-    For OpenMS based XIC workflow, file_pattern='chrom.mzML'.
+    This reads centroided LC-MS files from directory.
     '''
     print("Working on ~~ %s ~~ \n\n" %directory)
     return [os.path.join(directory, f) for f in os.listdir(directory) if file_pattern in f]
@@ -57,107 +80,6 @@ def create_export_folders(parameters, time_stamp):
     os.mkdir(os.path.join(parameters['outdir'], 'export'))
 
 
-# main workflow for `process`
-def process_project(list_input_files, parameters):
-    '''
-    list_input_files: Extracted ion chromatogram files.
-    parameters: dictionary of most parameters.
-    sample_registry include: 'status:mzml_parsing', 'status:eic', 'number_anchor_mz_pairs', 
-                'data_location', track_mzs: (mz, masstrack_id), 'sample_data': {}
-    '''
-    sample_registry = register_samples(list_input_files)
-    if parameters['database_mode'] == 'auto':
-        if len(list_input_files) <= parameters['project_sample_number_small']:
-            parameters['database_mode'] = 'memory'
-        elif len(list_input_files) <= parameters['project_sample_number_large']:
-            parameters['database_mode'] = 'ondisk'
-        else:
-            parameters['database_mode'] = 'ondisk'       # to implement 'split' 
-
-    # time_stamp is `month daay hour minute second``
-    time_stamp = ''.join([str(x) for x in time.localtime()[1:6]])
-    if parameters['database_mode'] == 'ondisk':
-        create_export_folders(parameters, time_stamp)
-        
-    # samples are processed to mass tracks (EICs) here
-    shared_dict = batch_EIC_from_samples_(sample_registry, parameters)
-    for sid, sam in sample_registry.items():
-        sam['status:mzml_parsing'], sam['status:eic'], sam['number_anchor_mz_pairs'
-                ], sam['data_location'], sam['track_mzs'], sam['sample_data'] = shared_dict[sid]
-        sam['name'] = os.path.basename(sam['input_file']).replace('.mzML', '')
-    
-    # print(sample_registry)
-    EE = ext_Experiment(sample_registry, parameters)
-    EE.process_all()
-
-    # export is separated, as it may be limited in some environments
-    if parameters['database_mode'] != 'ondisk':
-        create_export_folders(parameters, time_stamp)
-    EE.export_all()
-
-
-# -----------------------------------------------------------------------------
-# estimate_min_peak_height
-
-def estimate_min_peak_height(list_input_files, 
-            mz_tolerance_ppm=5, min_intensity=100, min_timepoints=5, min_peak_height=500,
-            num_files_to_use=3):
-    '''
-    return an estimated parameter for min peak_height as half of the min verified landmark peaks.
-    '''
-    estimated = []
-    if len(list_input_files) <= num_files_to_use:
-        selected = list_input_files
-    else:
-        selected = random.sample(list_input_files, num_files_to_use)
-    print("Estimating parameter for min peak_height based on ", selected)
-    for infile in selected:
-        try:
-            mz_landmarks, mode, min_peak_height_ = get_file_masstrack_stats(infile,
-                        mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height)
-                        # not all above parameters are used or relevant
-            estimated.append(min_peak_height_)
-        except:
-            print("Error in analyzing ", infile)
-    recommended = int(0.5 * np.median(estimated))
-    print("Estimated parameter for min peak_height is %d \n" %recommended)
-    return recommended
-
-def ext_estimate_min_peak_height(list_input_files, 
-            mz_tolerance_ppm=5, min_intensity=100, min_timepoints=5, min_peak_height=500,
-            num_files_to_use=3):
-    '''
-    return dict of
-    ion mode and
-    an estimated parameter for min peak_height as half of the min verified landmark peaks.
-
-    Extended estimate_min_peak_height for X-asari use.
-    '''
-    estimated, _ionmode = [], []
-    if len(list_input_files) <= num_files_to_use:
-        selected = list_input_files
-    else:
-        selected = random.sample(list_input_files, num_files_to_use)
-    print("Estimating parameter for min peak_height based on ", selected)
-    for infile in selected:
-        try:
-            mz_landmarks, mode, min_peak_height_ = get_file_masstrack_stats(infile,
-                        mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height)
-                        # not all above parameters are used or relevant
-            estimated.append(min_peak_height_)
-            _ionmode.append(mode)
-        except:
-            print("Error in analyzing ", infile)
-    recommended = int(0.5 * np.median(estimated))
-    if len(set(_ionmode)) > 1:
-        print("Error occured due to inconsistent ion mode." )
-        print(selected, _ionmode)
-        return None
-    else:
-        return {'mode': _ionmode[0], 'min_peak_height': recommended}
-
-
-
 # -----------------------------------------------------------------------------
 # Mass track (EIC) extraction, multi-core parralization via multiprocessing
 
@@ -168,15 +90,15 @@ def make_iter_parameters(sample_registry, parameters, shared_dict):
     [('sample_id', input_file, mode, mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, output_file, shared_dict), ...]
     '''
     iters = []
-    mz_tolerance_ppm = parameters['mz_tolerance']
+    mz_tolerance_ppm = parameters['mz_tolerance_ppm']
     min_intensity = parameters['min_intensity_threshold']
     min_timepoints = parameters['min_timepoints']
     min_peak_height = parameters['min_peak_height']
     for sample in sample_registry.values():
+        outfile = os.path.join(parameters['outdir'], 'pickle', os.path.basename(sample['input_file']).replace('.mzML', '')+'.pickle')
         iters.append(
             (sample['sample_id'], sample['input_file'], parameters['mode'], parameters['database_mode'],
-            mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height,
-            os.path.join(parameters['outdir'], 'pickle', os.path.basename(sample['input_file']).replace('.mzML', '')+'.pickle'),
+            mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, outfile,
             shared_dict
             )
         )
@@ -201,9 +123,15 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
                     mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, outfile, 
                     shared_dict):
     '''
-    Process infile. `shared_dict` is used to pass back information, 
-    sample_id: ('status:mzml_parsing', 'status:eic', number_anchor_mz_pairs, outfile, track_mzs, list_mass_tracks)
-    track_mzs are used later for aligning m/z tracks.
+    Process infile. 
+    `shared_dict` is used to pass back information, thus critical. Designed here as
+    sample_id: ('status:mzml_parsing', 'status:eic', outfile,
+                max_scan_number, list_scan_numbers, list_retention_time,
+                track_mzs,
+                number_anchor_mz_pairs, anchor_mz_pairs, 
+                dict({mass tracks}) )
+    track_mzs or anchor_mz_pairs are used later for aligning m/z tracks.
+    list of scans starts from 0. 
     '''
     new = {'sample_id': sample_id, 'input_file': infile, 'ion_mode': ion_mode,}
     list_mass_tracks = []
@@ -215,8 +143,7 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
                     min_intensity=min_intensity, 
                     min_timepoints=min_timepoints, 
                     min_peak_height=min_peak_height)
-        new['list_scan_numbers'] = xdict['rt_numbers']            # list of scans, starting from 0
-        new['list_retention_time'] = xdict['rt_times']        # full RT time points in sample
+        new['max_scan_number'] = max(xdict['rt_numbers'])
         ii = 0
         # already in ascending order of m/z from extract_massTracks_, get_thousandth_regions
         for track in xdict['tracks']:                         
@@ -236,18 +163,58 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
         new['number_anchor_mz_pairs'] = len(anchor_mz_pairs)
 
         if database_mode == 'ondisk':
-
-            shared_dict[new['sample_id']] = ('passed', 'passed', new['number_anchor_mz_pairs'], outfile, track_mzs, {})
+            shared_dict[new['sample_id']] = ('passed', 'passed', outfile,
+                                            new['max_scan_number'], xdict['rt_numbers'], xdict['rt_times'],
+                                            track_mzs,
+                                            new['number_anchor_mz_pairs'], anchor_mz_pairs,  
+                                            {} )  
             with open(outfile, 'wb') as f:
                 pickle.dump(new, f, pickle.HIGHEST_PROTOCOL)
 
         elif database_mode == 'memory':
-            shared_dict[new['sample_id']] = ('passed', 'passed', new['number_anchor_mz_pairs'], outfile, track_mzs, 
-                                            new)
+            shared_dict[new['sample_id']] = ('passed', 'passed', outfile,
+                                            new['max_scan_number'], xdict['rt_numbers'], xdict['rt_times'],
+                                            track_mzs,
+                                            new['number_anchor_mz_pairs'], anchor_mz_pairs,  
+                                            new )
 
-        print("Processed %s with %d mass tracks." %(os.path.basename(infile), ii))
+        print("Extracted %s to %d mass tracks." %(os.path.basename(infile), ii))
 
     except:
         # xml.etree.ElementTree.ParseError
-        shared_dict[new['sample_id']] = ('failed', '', 0, '', [], {})
+        shared_dict[new['sample_id']] = ('failed', '', '', 
+                                            0, [], [], [], 0, [], 
+                                            {})
         print("mzML processing error in sample %s, skipped." %infile)
+
+
+# -----------------------------------------------------------------------------
+# main workflow for `xic`
+# -----------------------------------------------------------------------------
+
+def process_xics(list_input_files, parameters):
+    '''
+    Get XICs (mass tracks) from a folder of centroid mzML files.
+    list_input_files: mzML files.
+    parameters: dictionary of most parameters.
+    '''
+    sample_registry = register_samples(list_input_files)
+    parameters['database_mode'] = 'ondisk'
+    time_stamp = ''.join([str(x) for x in time.localtime()[1:6]])
+    create_export_folders(parameters, time_stamp)
+    # samples are processed to mass tracks (EICs) here
+    shared_dict = batch_EIC_from_samples_(sample_registry, parameters)
+    print("XICs were stored as pickle objects under %s" %os.path.join(parameters['outdir'], 'pickle'))
+
+# -----------------------------------------------------------------------------
+# main workflow for `extract`
+# -----------------------------------------------------------------------------
+
+def get_mz_list(infile):
+    '''
+    Get a list of m/z valuies from infile, which is tab or comma delimited and has m/z in the first column.
+    Assuming first line as header and skip it.
+    '''
+    lines = open(infile).read().splitlines()[1:]
+    return [float(x.split('\t')[0].split(",")[0]) for x in lines]
+
